from operator import attrgetter
from typing import Dict, List, Iterable, Optional

import attr
from openpyxl import Workbook
from openpyxl.conftest import Worksheet
from openpyxl.utils import get_column_letter

from .models import Style, CellWithSize
from .utils import str_cell_range, style_and_merge_cell_range, style_single_cell


def xlsx_from_json(json_data: Dict, default_style: Style = None) -> Workbook:
    default_style = default_style or Style()

    wb = Workbook()
    sheet = wb.active

    start_row = json_data.get("start_row", 1)
    start_column = json_data.get("start_column", 1)

    filler = RowFiller(
        sheet, start_column, start_row, default_style, json_data.get("number_format")
    )
    row_positions = filler.fill(json_data.get("rows", []))

    adjuster = Adjuster(sheet)
    adjuster.adjust_columns(json_data.get("column_widths", []))
    adjuster.adjust_rows(json_data.get("rows", []), row_positions)

    return wb


@attr.s(auto_attribs=True)
class RowFiller:
    sheet: Worksheet
    start_column: int
    start_row: int
    default_style: Style
    number_format: Optional[str] = None

    def fill(self, rows_data: List[Dict]) -> Iterable[int]:
        current_row = self.start_row

        for row_data in rows_data:
            current_row += row_data.get("skip_rows", 0)
            start_column = self.start_column + row_data.get("skip_columns", 0)

            cells = self._fill_row(current_row, start_column, row_data.get("cells", []))
            row_height = max(map(attrgetter("height"), cells), default=1)

            yield current_row
            current_row += row_height

    def _fill_row(self, row: int, column: int, cells_data: List[Dict]) -> Iterable[CellWithSize]:
        current_column = column

        for cell_data in cells_data:
            cell = self._create_cell(row, current_column, cell_data)
            current_column += cell.width
            yield cell

    def _create_cell(self, row: int, column: int, cell_data: Dict) -> CellWithSize:
        cell = self.sheet.cell(row, column)

        value = cell_data["value"]
        cell.value = value

        if isinstance(value, float) and self.number_format:
            cell.number_format = self.number_format

        width = cell_data.get("width", 1)
        height = cell_data.get("height", 1)

        style = Style.from_json(cell_data.get("style", {}), self.default_style)

        if width == 1 and height == 1:
            style_single_cell(cell, style)
        else:
            cell_range = str_cell_range(column, row, column + width - 1, row + height - 1)
            style_and_merge_cell_range(self.sheet, cell_range, style)

        return CellWithSize(cell, width, height)


@attr.s(auto_attribs=True)
class Adjuster:
    sheet: Worksheet

    def adjust_columns(self, columns: Iterable[Dict]) -> None:
        for column_data in columns:
            if "column_number" in column_data:
                column_letter = get_column_letter(column_data["column_number"])
            elif "column_letter" in column_data:
                column_letter = column_data["column_letter"]
            else:
                raise ValueError("No column provided.")

            self.sheet.column_dimensions[column_letter].width = column_data["width"]

    def adjust_rows(self, rows: Iterable[Dict], row_positions: Iterable[int]) -> None:
        for row_data, row_position in zip(rows, row_positions):
            if "row_height" in row_data:
                self.sheet.row_dimensions[row_position].height = row_data["row_height"]



import pytest
from openpyxl import Workbook
from openpyxl.styles import Font
from openpyxl.utils import get_column_letter

from xlsx_from_json import xlsx_from_json, Style


@pytest.fixture()
def default_style():
    return Style(font=Font(bold=True))


@pytest.fixture()
def json_data_with_start_row_and_multiple_cells():
    return {
        "rows": [
            {
                "cells": [
                    {
                        "value": "1x2",
                        "height": 2
                    }
                ]
            },
            {
                "cells": [
                    {
                        "value": "2x1",
                        "width": 2
                    },
                    {
                        "value": "1x1"
                    }
                ]
            }
        ],
        "start_row": 2
    }


def test_sheet_has_values():
    wb = xlsx_from_json({"rows": [{"cells": [{"value": "op"}]}]})
    sheet = wb.active
    assert sheet.cell(1, 1).value == "op"


def test_cell_column_equals_to_start_column():
    wb = xlsx_from_json({"rows": [{"cells": [{"value": "op"}]}], "start_column": 3})
    sheet = wb.active
    assert sheet.cell(1, 3).value == "op"


def test_cell_has_font():
    wb = xlsx_from_json({
        "rows": [
            {
                "cells": [
                    {
                        "value": "op",
                        "style": {"font": {"size": 12, "name": "Times New Roman"}}
                    }
                ]
            }
        ]
    })
    sheet = wb.active
    cell = sheet.cell(1, 1)
    assert cell.font.name == "Times New Roman"
    assert cell.font.size == 12


def test_cell_has_font_bold_from_default_style(default_style):
    wb = xlsx_from_json({"rows": [{"cells": [{"value": "op"}]}]}, default_style)
    sheet = wb.active
    cell = sheet.cell(1, 1)
    assert cell.font.bold


def test_cell_has_border():
    wb = xlsx_from_json({
        "rows": [
            {
                "cells": [
                    {
                        "value": "op",
                        "style": {"border": {"bottom": {"border_style": "medium", "color": "00000000"}}}
                    }
                ]
            }
        ]
    })
    sheet = wb.active
    cell = sheet.cell(1, 1)
    assert cell.border.bottom.border_style == "medium"
    assert cell.border.bottom.color.rgb == "00000000"


def test_cell_has_alignment():
    wb = xlsx_from_json({
        "rows": [
            {
                "cells": [
                    {
                        "value": "op",
                        "style": {"alignment": {"horizontal": "center"}}
                    }
                ]
            }
        ]
    })
    sheet = wb.active
    cell = sheet.cell(1, 1)
    assert cell.alignment.horizontal == "center"


def test_cell_has_fill():
    wb = xlsx_from_json({
        "rows": [
            {
                "cells": [
                    {
                        "value": "op",
                        "style": {"fill": {"patternType": "gray125"}}
                    }
                ]
            }
        ]
    })
    sheet = wb.active
    cell = sheet.cell(1, 1)
    assert cell.fill.patternType == "gray125"


def test_sized_cell_is_rendered_as_merged_cells_and_style_set():
    workbook: Workbook = xlsx_from_json({
        "rows": [
            {
                "cells": [
                    {
                        "value": "Sample text",
                        "width": 5,
                        "height": 2,
                        "style": {
                            "font": {
                                "name": "Times New Roman",
                                "size": 12
                            },
                            "border": {
                                "bottom": {
                                    "border_style": "medium",
                                    "color": "FFFFFFFF"
                                }
                            }
                        }
                    }
                ]
            }
        ]
    })
    sheet = workbook.active
    assert sheet.cell(2, 5).border.bottom.border_style == "medium"


def test_sheet_fill_starts_with_start_row(json_data_with_start_row_and_multiple_cells):
    workbook: Workbook = xlsx_from_json(json_data_with_start_row_and_multiple_cells)
    sheet = workbook.active
    assert sheet.cell(row=2, column=1).value == "1x2"


def test_sheet_has_two_rows(json_data_with_start_row_and_multiple_cells):
    workbook: Workbook = xlsx_from_json(json_data_with_start_row_and_multiple_cells)
    sheet = workbook.active
    assert sheet.cell(row=2, column=1).value == "1x2"
    assert sheet.cell(row=4, column=1).value == "2x1"


def test_sheet_has_two_columns(json_data_with_start_row_and_multiple_cells):
    workbook: Workbook = xlsx_from_json(json_data_with_start_row_and_multiple_cells)
    sheet = workbook.active
    assert sheet.cell(row=4, column=1).value == "2x1"
    assert sheet.cell(row=4, column=3).value == "1x1"


def test_empty_row_render():
    workbook: Workbook = xlsx_from_json({
        "rows": [
            {"cells": []},
            {"cells": [{"value": "op"}]},
            {},
            {"cells": [{"value": "op"}]},
        ]
    })
    sheet = workbook.active
    assert sheet.cell(2, 1).value == "op"
    assert sheet.cell(4, 1).value == "op"


def test_row_and_column_skip():
    workbook: Workbook = xlsx_from_json({
        "rows": [
            {"cells": [{"value": "op"}], "skip_rows": 2, "skip_columns": 3}
        ]
    })
    sheet = workbook.active
    assert sheet.cell(3, 4).value == "op"


def test_row_and_column_sizing():
    workbook: Workbook = xlsx_from_json({
        "rows": [
            {"cells": [{"value": "op"}], "row_height": 10},
        ],
        "column_widths": [
            {
                "width": 10,
                "column_number": 1
            },
            {
                "width": 20,
                "column_letter": "B"
            }
        ]
    })
    sheet = workbook.active
    assert sheet.row_dimensions[1].height == 10
    assert sheet.column_dimensions[get_column_letter(1)].width == 10
    assert sheet.column_dimensions["B"].width == 20


def test_float_formatting():
    workbook: Workbook = xlsx_from_json({
        "rows": [
            {"cells": [{"value": 1.256}]},
        ],
        "number_format": "0.00"
    })
    sheet = workbook.active
    assert sheet.cell(1, 1).number_format == "0.00"
